from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class Vec2I64(_message.Message):
    __slots__ = ("x", "y")
    X_FIELD_NUMBER: _ClassVar[int]
    Y_FIELD_NUMBER: _ClassVar[int]
    x: int
    y: int
    def __init__(self, x: _Optional[int] = ..., y: _Optional[int] = ...) -> None: ...

class Vec2F32(_message.Message):
    __slots__ = ("x", "y")
    X_FIELD_NUMBER: _ClassVar[int]
    Y_FIELD_NUMBER: _ClassVar[int]
    x: float
    y: float
    def __init__(self, x: _Optional[float] = ..., y: _Optional[float] = ...) -> None: ...

class Vec2F64(_message.Message):
    __slots__ = ("x", "y")
    X_FIELD_NUMBER: _ClassVar[int]
    Y_FIELD_NUMBER: _ClassVar[int]
    x: float
    y: float
    def __init__(self, x: _Optional[float] = ..., y: _Optional[float] = ...) -> None: ...

class Vec3I64(_message.Message):
    __slots__ = ("x", "y", "z")
    X_FIELD_NUMBER: _ClassVar[int]
    Y_FIELD_NUMBER: _ClassVar[int]
    Z_FIELD_NUMBER: _ClassVar[int]
    x: int
    y: int
    z: int
    def __init__(self, x: _Optional[int] = ..., y: _Optional[int] = ..., z: _Optional[int] = ...) -> None: ...

class Vec3F32(_message.Message):
    __slots__ = ("x", "y", "z")
    X_FIELD_NUMBER: _ClassVar[int]
    Y_FIELD_NUMBER: _ClassVar[int]
    Z_FIELD_NUMBER: _ClassVar[int]
    x: float
    y: float
    z: float
    def __init__(self, x: _Optional[float] = ..., y: _Optional[float] = ..., z: _Optional[float] = ...) -> None: ...

class Vec3F64(_message.Message):
    __slots__ = ("x", "y", "z")
    X_FIELD_NUMBER: _ClassVar[int]
    Y_FIELD_NUMBER: _ClassVar[int]
    Z_FIELD_NUMBER: _ClassVar[int]
    x: float
    y: float
    z: float
    def __init__(self, x: _Optional[float] = ..., y: _Optional[float] = ..., z: _Optional[float] = ...) -> None: ...

class Mat2F32(_message.Message):
    __slots__ = ("col_0", "col_1")
    COL_0_FIELD_NUMBER: _ClassVar[int]
    COL_1_FIELD_NUMBER: _ClassVar[int]
    col_0: Vec2F32
    col_1: Vec2F32
    def __init__(self, col_0: _Optional[_Union[Vec2F32, _Mapping]] = ..., col_1: _Optional[_Union[Vec2F32, _Mapping]] = ...) -> None: ...

class Mat2F64(_message.Message):
    __slots__ = ("col_0", "col_1")
    COL_0_FIELD_NUMBER: _ClassVar[int]
    COL_1_FIELD_NUMBER: _ClassVar[int]
    col_0: Vec2F64
    col_1: Vec2F64
    def __init__(self, col_0: _Optional[_Union[Vec2F64, _Mapping]] = ..., col_1: _Optional[_Union[Vec2F64, _Mapping]] = ...) -> None: ...

class Mat3F32(_message.Message):
    __slots__ = ("col_0", "col_1", "col_2")
    COL_0_FIELD_NUMBER: _ClassVar[int]
    COL_1_FIELD_NUMBER: _ClassVar[int]
    COL_2_FIELD_NUMBER: _ClassVar[int]
    col_0: Vec3F32
    col_1: Vec3F32
    col_2: Vec3F32
    def __init__(self, col_0: _Optional[_Union[Vec3F32, _Mapping]] = ..., col_1: _Optional[_Union[Vec3F32, _Mapping]] = ..., col_2: _Optional[_Union[Vec3F32, _Mapping]] = ...) -> None: ...

class Mat3F64(_message.Message):
    __slots__ = ("col_0", "col_1", "col_2")
    COL_0_FIELD_NUMBER: _ClassVar[int]
    COL_1_FIELD_NUMBER: _ClassVar[int]
    COL_2_FIELD_NUMBER: _ClassVar[int]
    col_0: Vec3F64
    col_1: Vec3F64
    col_2: Vec3F64
    def __init__(self, col_0: _Optional[_Union[Vec3F64, _Mapping]] = ..., col_1: _Optional[_Union[Vec3F64, _Mapping]] = ..., col_2: _Optional[_Union[Vec3F64, _Mapping]] = ...) -> None: ...

class VecXF32(_message.Message):
    __slots__ = ("data",)
    DATA_FIELD_NUMBER: _ClassVar[int]
    data: _containers.RepeatedScalarFieldContainer[float]
    def __init__(self, data: _Optional[_Iterable[float]] = ...) -> None: ...

class VecXF64(_message.Message):
    __slots__ = ("data",)
    DATA_FIELD_NUMBER: _ClassVar[int]
    data: _containers.RepeatedScalarFieldContainer[float]
    def __init__(self, data: _Optional[_Iterable[float]] = ...) -> None: ...

class MatXF32(_message.Message):
    __slots__ = ("rows", "cols", "data")
    ROWS_FIELD_NUMBER: _ClassVar[int]
    COLS_FIELD_NUMBER: _ClassVar[int]
    DATA_FIELD_NUMBER: _ClassVar[int]
    rows: int
    cols: int
    data: _containers.RepeatedScalarFieldContainer[float]
    def __init__(self, rows: _Optional[int] = ..., cols: _Optional[int] = ..., data: _Optional[_Iterable[float]] = ...) -> None: ...

class MatXF64(_message.Message):
    __slots__ = ("rows", "cols", "data")
    ROWS_FIELD_NUMBER: _ClassVar[int]
    COLS_FIELD_NUMBER: _ClassVar[int]
    DATA_FIELD_NUMBER: _ClassVar[int]
    rows: int
    cols: int
    data: _containers.RepeatedScalarFieldContainer[float]
    def __init__(self, rows: _Optional[int] = ..., cols: _Optional[int] = ..., data: _Optional[_Iterable[float]] = ...) -> None: ...



from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional

DESCRIPTOR: _descriptor.FileDescriptor

class Timestamp(_message.Message):
    __slots__ = ("stamp", "clock_name", "semantics")
    STAMP_FIELD_NUMBER: _ClassVar[int]
    CLOCK_NAME_FIELD_NUMBER: _ClassVar[int]
    SEMANTICS_FIELD_NUMBER: _ClassVar[int]
    stamp: float
    clock_name: str
    semantics: str
    def __init__(self, stamp: _Optional[float] = ..., clock_name: _Optional[str] = ..., semantics: _Optional[str] = ...) -> None: ...



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: farm_ng/core/event.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from farm_ng.core import uri_pb2 as farm__ng_dot_core_dot_uri__pb2
from farm_ng.core import timestamp_pb2 as farm__ng_dot_core_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x18\x66\x61rm_ng/core/event.proto\x12\x12\x66\x61rm_ng.core.proto\x1a\x16\x66\x61rm_ng/core/uri.proto\x1a\x1c\x66\x61rm_ng/core/timestamp.proto\"\x8a\x01\n\x05\x45vent\x12$\n\x03uri\x18\x01 \x01(\x0b\x32\x17.farm_ng.core.proto.Uri\x12\x31\n\ntimestamps\x18\x02 \x03(\x0b\x32\x1d.farm_ng.core.proto.Timestamp\x12\x16\n\x0epayload_length\x18\x03 \x01(\x03\x12\x10\n\x08sequence\x18\x04 \x01(\x03\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'farm_ng.core.event_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_EVENT']._serialized_start=103
  _globals['_EVENT']._serialized_end=241
# @@protoc_insertion_point(module_scope)



from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional

DESCRIPTOR: _descriptor.FileDescriptor

class Color(_message.Message):
    __slots__ = ("r", "g", "b", "a")
    R_FIELD_NUMBER: _ClassVar[int]
    G_FIELD_NUMBER: _ClassVar[int]
    B_FIELD_NUMBER: _ClassVar[int]
    A_FIELD_NUMBER: _ClassVar[int]
    r: float
    g: float
    b: float
    a: float
    def __init__(self, r: _Optional[float] = ..., g: _Optional[float] = ..., b: _Optional[float] = ..., a: _Optional[float] = ...) -> None: ...



from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class ImageSize(_message.Message):
    __slots__ = ("width", "height")
    WIDTH_FIELD_NUMBER: _ClassVar[int]
    HEIGHT_FIELD_NUMBER: _ClassVar[int]
    width: int
    height: int
    def __init__(self, width: _Optional[int] = ..., height: _Optional[int] = ...) -> None: ...

class ImageLayout(_message.Message):
    __slots__ = ("size", "pitch_bytes")
    SIZE_FIELD_NUMBER: _ClassVar[int]
    PITCH_BYTES_FIELD_NUMBER: _ClassVar[int]
    size: ImageSize
    pitch_bytes: int
    def __init__(self, size: _Optional[_Union[ImageSize, _Mapping]] = ..., pitch_bytes: _Optional[int] = ...) -> None: ...

class PixelFormat(_message.Message):
    __slots__ = ("number_type", "num_components", "num_bytes_per_component")
    NUMBER_TYPE_FIELD_NUMBER: _ClassVar[int]
    NUM_COMPONENTS_FIELD_NUMBER: _ClassVar[int]
    NUM_BYTES_PER_COMPONENT_FIELD_NUMBER: _ClassVar[int]
    number_type: str
    num_components: int
    num_bytes_per_component: int
    def __init__(self, number_type: _Optional[str] = ..., num_components: _Optional[int] = ..., num_bytes_per_component: _Optional[int] = ...) -> None: ...

class DynImage(_message.Message):
    __slots__ = ("layout", "pixel_format", "data")
    LAYOUT_FIELD_NUMBER: _ClassVar[int]
    PIXEL_FORMAT_FIELD_NUMBER: _ClassVar[int]
    DATA_FIELD_NUMBER: _ClassVar[int]
    layout: ImageLayout
    pixel_format: PixelFormat
    data: bytes
    def __init__(self, layout: _Optional[_Union[ImageLayout, _Mapping]] = ..., pixel_format: _Optional[_Union[PixelFormat, _Mapping]] = ..., data: _Optional[bytes] = ...) -> None: ...



from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class StructExample1(_message.Message):
    __slots__ = ("integer", "f")
    INTEGER_FIELD_NUMBER: _ClassVar[int]
    F_FIELD_NUMBER: _ClassVar[int]
    integer: float
    f: float
    def __init__(self, integer: _Optional[float] = ..., f: _Optional[float] = ...) -> None: ...

class StructExample2(_message.Message):
    __slots__ = ("name", "ex1")
    NAME_FIELD_NUMBER: _ClassVar[int]
    EX1_FIELD_NUMBER: _ClassVar[int]
    name: str
    ex1: StructExample1
    def __init__(self, name: _Optional[str] = ..., ex1: _Optional[_Union[StructExample1, _Mapping]] = ...) -> None: ...



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from farm_ng.core import pose_pb2

# Usage example:
# pose_tree = ... # Load your PoseTree proto object here.
# result = find_poses_by_frames(pose_tree, "base_link", "wheel_link")
# print(result)

# NOTE: pep8-naming throws an error here because names should be snake_case.


def find_pose(
    pose_tree: pose_pb2.PoseTree,
    frame_a: str,
    frame_b: str,
) -> pose_pb2.Pose | None:
    """Find all poses in a PoseTree with a given frame_A and frame_B.

    If frames are inverted, returns the inverse transform.

    Args:
        pose_tree (PoseTree): The PoseTree proto object to search within.
        frame_a (str): The desired frame_a value.
        frame_b (str): The desired frame_b value.

    Returns:
        Pose: Pose objects that match the criteria.
    """
    pose: pose_pb2.Pose
    for pose in pose_tree.poses:
        if pose.frame_a == frame_a and pose.frame_b == frame_b:
            return pose
        if pose.frame_a == frame_b and pose.frame_b == frame_a:
            print("NOT IMPLEMENTED, please swap your frames")
            return None
    return None



import os
import sys
from pathlib import Path

from farm_ng.package.commands import (
    BuildProtosDevelop,
    BuildProtosEggInfo,
    BuildProtosInstall,
    CleanFilesCommand,
)
from pybind11.setup_helpers import ParallelCompile, Pybind11Extension, build_ext
from setuptools import setup

__version__ = "2.3.1"


platform_cxx_flags = []

if sys.platform.startswith("darwin"):
    print("Running on macOS")
    platform_cxx_flags = [
        "-std=c++20",
        "-mmacosx-version-min=11.0",
        "-D_LIBCPP_DISABLE_DEPRECATION_WARNINGS=1",
    ]
    os.environ["MACOSX_DEPLOYMENT_TARGET"] = "11.0"

elif sys.platform.startswith("linux"):
    print("Running on Linux")
    platform_cxx_flags = [
        "-std=gnu++17",
        "-fconcepts",
    ]
else:
    print("Running on another operating system")

PROTO_ROOT: str = "protos"
PACKAGE_ROOT: str = "py"


BuildProtosDevelop.user_options.append(("proto-root=", None, PROTO_ROOT))
BuildProtosDevelop.user_options.append(("package-root=", None, PACKAGE_ROOT))

BuildProtosInstall.user_options.append(("proto-root=", None, PROTO_ROOT))
BuildProtosInstall.user_options.append(("package-root=", None, PACKAGE_ROOT))

BuildProtosEggInfo.user_options.append(("proto-root=", None, PROTO_ROOT))
BuildProtosEggInfo.user_options.append(("package-root=", None, PACKAGE_ROOT))

CleanFilesCommand.user_options.append(("package-root=", None, PACKAGE_ROOT))


# Optional multithreaded build
ParallelCompile("CMAKE_BUILD_PARALLEL_LEVEL", default=4).install()

source_path = Path(__file__).resolve()
source_dir = source_path.parent


# The main interface is through Pybind11Extension.
# * You can add cxx_std=11/14/17, and then build_ext can be removed.
# * You can set include_pybind11=false to add the include directory yourself,
#   say from a submodule.
#
# Note:
#   Sort input source files if you glob sources to ensure bit-for-bit
#   reproducible builds (https://github.com/pybind/python_example/pull/53)


ext_modules = [
    Pybind11Extension(
        "farm_ng_core_pybind",
        [
            # farm-ng-control
            "py/pybind/farm_ng_core_pybind.cpp",
            "py/pybind/linalg_pybind.cpp",
            "py/pybind/lie_pybind.cpp",
            # fmt
            "cpp/thirdparty/fmt/src/format.cc",
            "cpp/thirdparty/fmt/src/os.cc",
            # farm-ng-core
            "cpp/farm_ng/core/logging/backtrace.cpp",
            "cpp/farm_ng/core/logging/expected.cpp",
            "cpp/farm_ng/core/logging/format.cpp",
            "cpp/farm_ng/core/logging/logger.cpp",
        ],
        define_macros=[("VERSION_INFO", __version__)],
        extra_compile_args=[
            *platform_cxx_flags,
            "-Wall",
            "-Wextra",
            "-Werror",
            "-Wno-unknown-warning-option",
            "-Wno-unused-parameter",
            "-Wno-missing-field-initializers",
            "-Wno-unused-but-set-variable",
            "-Wno-unused-variable",
            "-Wno-unused-function",
            "-Wno-maybe-uninitialized",
        ],
        include_dirs=[
            source_dir / "cpp",
            source_dir / "cpp/thirdparty/expected/include",
            source_dir / "cpp/thirdparty/farm_pp/include",
            source_dir / "cpp/thirdparty/fmt/include",
            source_dir / "cpp/thirdparty/eigen",
        ],
    ),
]

setup(
    ext_modules=ext_modules,
    cmdclass={
        "install": BuildProtosInstall,
        "develop": BuildProtosDevelop,
        "egg_info": BuildProtosEggInfo,
        "clean": CleanFilesCommand,
        "build_ext": build_ext,
    },
)



from farm_ng.core import event_service_pb2 as _event_service_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class RecorderProfile(_message.Message):
    __slots__ = ("name", "subscriptions")
    NAME_FIELD_NUMBER: _ClassVar[int]
    SUBSCRIPTIONS_FIELD_NUMBER: _ClassVar[int]
    name: str
    subscriptions: _containers.RepeatedCompositeFieldContainer[_event_service_pb2.SubscribeRequest]
    def __init__(self, name: _Optional[str] = ..., subscriptions: _Optional[_Iterable[_Union[_event_service_pb2.SubscribeRequest, _Mapping]]] = ...) -> None: ...



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: farm_ng/core/examples.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1b\x66\x61rm_ng/core/examples.proto\x12\x12\x66\x61rm_ng.core.proto\",\n\x0eStructExample1\x12\x0f\n\x07integer\x18\x01 \x01(\x01\x12\t\n\x01\x66\x18\x02 \x01(\x02\"O\n\x0eStructExample2\x12\x0c\n\x04name\x18\x01 \x01(\t\x12/\n\x03\x65x1\x18\x02 \x01(\x0b\x32\".farm_ng.core.proto.StructExample1b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'farm_ng.core.examples_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_STRUCTEXAMPLE1']._serialized_start=51
  _globals['_STRUCTEXAMPLE1']._serialized_end=95
  _globals['_STRUCTEXAMPLE2']._serialized_start=97
  _globals['_STRUCTEXAMPLE2']._serialized_end=176
# @@protoc_insertion_point(module_scope)



"""
Usage:
python -m farm_ng.core.event_service_tool uris config.json

This will print a config file will all the uris currently being published by the config

python -m farm_ng.core.event_service_tool config-gen config2.json

This print a sample config file with a single service called test_service, and an example
recorder service configuration (note with no port numbers so scripts no not to connect)

Runs a single service from the config file

python -m farm_ng.core.event_service_tool launch1 --service-config config.json --service-name test_service

"""
import argparse
import asyncio
import sys

from farm_ng.core.event_client import EventClient
from farm_ng.core.event_service import add_service_parser, load_service_config
from farm_ng.core.event_service_pb2 import (
    EventServiceConfig,
    EventServiceConfigList,
    SubscribeRequest,
)
from farm_ng.core.events_file_writer import proto_to_json_file
from farm_ng.core.uri_pb2 import Uri
from google.protobuf.json_format import MessageToJson


async def config_gen_command(args):
    proto_to_json_file(
        args.output,
        EventServiceConfigList(
            configs=[
                EventServiceConfig(
                    name=args.name,
                    host=args.host,
                    port=args.port,
                    subscriptions=[
                        SubscribeRequest(
                            uri=Uri(path="/test", query="service_name=bar"),
                            every_n=1,
                        ),
                    ],
                    args=["--my-arg=my-value", "foo"],
                    log_level=EventServiceConfig.LogLevel.INFO,
                ),
                EventServiceConfig(
                    name="record_default",
                    subscriptions=[
                        SubscribeRequest(
                            uri=Uri(path="*", query="service_name={args.name}"),
                            every_n=1,
                        ),
                    ],
                ),
            ],
        ),
    )


async def list_uri(config: EventServiceConfig):
    print("connecting to", config)
    client = EventClient(config)
    uris = await client.list_uris()
    del config.uris[:]
    config.uris.extend(uris)
    return config


async def uris_command(args):
    configs, service_config = load_service_config(args)
    config = await list_uri(service_config)
    print(MessageToJson(config))


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    sub_parsers = parser.add_subparsers()
    config_gen_parser = sub_parsers.add_parser("config-gen")
    config_gen_parser.add_argument("output")
    config_gen_parser.add_argument("--host", default="localhost")
    config_gen_parser.add_argument("--name", default="test_service")
    config_gen_parser.add_argument("--port", default=5001, type=int)
    config_gen_parser.add_argument("--subscription", default="/test", type=str)

    config_gen_parser.set_defaults(func=config_gen_command)

    uris_parser = sub_parsers.add_parser("uris")
    add_service_parser(uris_parser)
    uris_parser.set_defaults(func=uris_command)

    args = parser.parse_args()
    if hasattr(args, "func"):
        asyncio.run(args.func(args))
    else:
        parser.print_help()
        sys.exit(1)



from farm_ng.core import uri_pb2 as _uri_pb2
from farm_ng.core import timestamp_pb2 as _timestamp_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class Event(_message.Message):
    __slots__ = ("uri", "timestamps", "payload_length", "sequence")
    URI_FIELD_NUMBER: _ClassVar[int]
    TIMESTAMPS_FIELD_NUMBER: _ClassVar[int]
    PAYLOAD_LENGTH_FIELD_NUMBER: _ClassVar[int]
    SEQUENCE_FIELD_NUMBER: _ClassVar[int]
    uri: _uri_pb2.Uri
    timestamps: _containers.RepeatedCompositeFieldContainer[_timestamp_pb2.Timestamp]
    payload_length: int
    sequence: int
    def __init__(self, uri: _Optional[_Union[_uri_pb2.Uri, _Mapping]] = ..., timestamps: _Optional[_Iterable[_Union[_timestamp_pb2.Timestamp, _Mapping]]] = ..., payload_length: _Optional[int] = ..., sequence: _Optional[int] = ...) -> None: ...



from __future__ import annotations

from collections import deque

from google.protobuf.struct_pb2 import Struct

__all__ = ["EventServiceHealthMetrics"]


class Singleton(type):
    _instances: dict[type, type] = {}  # noqa: RUF012

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]


class EventServiceHealthMetrics(metaclass=Singleton):
    MAX_QUEUE_SIZE: int = 50  # size of the circular buffer
    MIN_NUM_ELEMENTS: int = 2  # minimum number of elements to compute the metrics

    def __init__(self) -> None:
        # initialize the data with the default values
        self.data = Struct()
        self.stamps_buffer: dict[str, deque[float]] = {}

    # public methods

    def update_data(self, uri_path: str, value: float) -> None:
        if uri_path not in self.data:
            self.data[uri_path] = 0.0

        self.data.update({uri_path: value})

    def get(self, uri_path: str):
        if uri_path not in self.data:
            return 0

        return self.data[uri_path]

    def update_stamps(self, uri_path: str, stamp: float) -> None:
        if uri_path not in self.stamps_buffer:
            self.stamps_buffer[uri_path] = deque(maxlen=self.MAX_QUEUE_SIZE)

        self.stamps_buffer[uri_path].append(stamp)

    def compute_data(self) -> Struct:
        """Compute the metrics data that requires time based computation."""

        # iterate over the stamps buffer
        for topic, stamps in self.stamps_buffer.items():
            # compute rate
            self.data[f"{topic}/rate"] = self._compute_avg_rate(stamps)
            # compute rate iter, need to be checked
            # self.data[f"{topic}/rate"] = self._compute_avg_rate_iter(stamps)

        return self.data

    def reset_data(self) -> None:
        self.data = Struct()
        self.stamps_buffer = {}

    # private methods

    def _compute_avg_rate_iter(self, stamps: deque[float]) -> float:
        if len(stamps) < self.MIN_NUM_ELEMENTS:
            return 0.0

        latencies: list[float] = [
            stamps[i] - stamps[i - 1] for i in range(1, len(stamps))
        ]

        return len(latencies) / sum(latencies)

    def _compute_avg_rate(self, stamps: deque[float]) -> float:
        if len(stamps) < self.MIN_NUM_ELEMENTS:
            return 0.0

        duration = stamps[-1] - stamps[0]
        return len(stamps) / duration if duration > 0 else 0.0
